import re
import subprocess
from collections import namedtuple
from enum import Enum, Flag, auto

import requests
from emojis import emojis

from wommit.apihandler import ApiHandler
from wommit.configman import Configurator
from wommit.utils import utils
from wommit.msgtool import MsgTool, RegVal
from wommit.utils.utils import get_repo_info
from prompt_toolkit.shortcuts import ProgressBar

CommitAppend = namedtuple("CommitAppend", ["type", "message"])

class LogVersion(Flag):
    BASIC = auto()
    RELEASE = auto()

class ChangelogGenerator:

    def __init__(self, logver:LogVersion=LogVersion.BASIC):
        self.api_url = "https://api.github.com/graphql"
        self.token = ApiHandler().get_token()
        self.data = Configurator().load_data()
        self.data = utils.load_data()
        self.dont_include = ['Merge', 'bump', 'Update']
        self.logver = logver

    def _r_func(self, query, variables, token):
        return requests.post(
            self.api_url,
            json={"query": query, "variables": variables},
            headers={"Authorization": "token {}".format(token)},
        )

    def generate_changelog(self, pre_release=True):
        commits = self._commits_since_release(pre_release)
        typedict = self._gen_type_list()
        typedict['others'] = []
        # print(commits)
        map = self.data['emoji_map']
        inv_map = {v: k for k, v in map.items()}

        dont_include_joined = "|".join(self.dont_include)

        msgtool = MsgTool(self.data)

        # if sys.stdout.isatty():
        # with ProgressBar() as pb:

        for commit in commits:
            msgc = msgtool.check_message(commit)
            good_commits = []

            if msgc:

                # Change these three to enums at some point.
                breaking, typeman, subject = msgtool.get_reg_vals(msgc,
                                                                  [RegVal.breaking, RegVal.type, RegVal.subject])

                for group in [breaking, typeman]:
                    if group:
                        emoji_text = emojis.decode(group)
                        if emoji_text in inv_map:
                            emoji_val = inv_map[emoji_text]
                            good_commits.append(
                                CommitAppend(emoji_val, subject))  # Adds a new commit message under type list.
                        else:
                            print("NON EMOJI VAL SPOTED")
                            good_commits.append(CommitAppend('others', commit))
            else:
                good_commits.append(CommitAppend('others', commit))

            for val in good_commits:
                if not re.match(f'^({dont_include_joined})', val.message):
                    typedict[val.type].append(val.message)

        # print(typedict)
        changelog = self._changelog_formatter(typedict)
        return changelog

    def _changelog_formatter(self, values):
        retstring = ""
        for typeman, commits in values.items():
            if commits:
                commits = filter(lambda x: bool(x), commits)
                try:
                    emoji =  emojis.encode(self.data['emoji_map'][typeman])
                except KeyError:
                    emoji = emojis.encode(":japanese_goblin:")
                typeman = (typeman + "s" if not typeman[-1] == "s" else typeman).lower()
                seperator = "\n\n - "
                commits_ordered = " - " + seperator.join(commits)
                # commits_ordered = f" - {seperator.join(commits)}"
                if self.logver == LogVersion.BASIC:
                    retstring += f"{emoji} {typeman}:\n\n{commits_ordered}\n\n"
                elif self.logver == LogVersion.RELEASE:
                    retstring += f"<details open>\n<summary>{emoji} {typeman}</summary>\n\n{commits_ordered}\n\n</details><br>\n<br>\n"
                else:
                    print("Unsupported log version")
                    return

        return retstring

    def _gen_type_list(self):
        # types = [a['value'] for a in self.data['types']]
        types = [a for a in self.data['emoji_map'].keys()]
        return {x: [] for x in types}

    def _commits_since_release(self, pre_release:bool):

        data = ApiHandler().get_releases(5)

        try:
            if pre_release:
                last_tag = data[0]['node']['tagName']
                args = f'git log {last_tag}..HEAD --oneline'.split(' ')
            else:
                last_tag = data[0]['node']['tagName']
                earlier_tag = data[1]['node']['tagName']
                args = f'git log {earlier_tag}..{last_tag} --oneline'.split(' ')
        except (IndexError, KeyError) as err: # If there are no earlier releases, get all commits.
            args = f'git log --oneline'.split(' ')

        args.append('--pretty=format:%s (%h)')
        #try:
        moreargs = {'text': True, 'check': True, 'capture_output': True, 'encoding': 'utf-8'}
        call = subprocess.run(args, **moreargs)
        out = call.stdout
        #except subprocess.CalledProcessError as err:
            #print(err.stderr)
            #out =err.stderr

        retlist = (out).split('\n')
        return retlist






if __name__ == '__main__':
    ChangelogGenerator().generate_changelog()


import re
from enum import Enum
from typing import Union

import emojis
from prompt_toolkit.document import Document

from wommit.utils.utils import load_data
from prompt_toolkit.validation import Validator, ValidationError


class RegVal(Enum):
    # Which group of the regex contains what part.
    breaking = 1
    type = 2
    exclamation = 3
    scope = 4
    subject = 5

class MsgTool:

    def __init__(self, data=None):
        if not data:
            data = load_data()
        self.types = [a["value"] for a in data["types"]]
        self.emomap = data["emoji_map"]
        matchedemos = [b for (a,b) in self.emomap.items() if a in self.types]
        self.emojis = [emojis.encode(a) for a in matchedemos]



    def check_and_convert(self, input: Union[str, dict]):

        if isinstance(input, dict):
            input = self.msg_dict_to_str(input)

        m = self.check_message(input)
        if m:
            return self._convert_message(input, m)

    def check_message(self, message) -> re.Match:
        """
        Checks if the commit message conforms to
        the conventional commit standard.
        :param message: commit message to check
        :param types: mandatory types for repo. if empty, all types are allowed.
        :return: True if the message is valid, False otherwise.
        """
        # if types:
        #     typestr = "|".join(types)
        if self.types:
            typestr = "|".join(self.types + self.emojis)
        else:
            typestr = ".+" # allow anything as type
        boom = emojis.emojis.db.get_emoji_by_alias("boom").emoji

        ##
        reg_s = r"^({boom})?({typestr})(!?)(\([a-z,]+\))? ([\s\S]+\n?)".format(typestr=typestr, boom=boom) # Double { used to escape.
        ##

        m = re.match(reg_s, message)
        if not m:
            mergematch = re.match(r"^Merge branch [\w] into [\w]", message)
            return mergematch

        return m

    def get_reg_vals(self, match: re.Match, vals:[RegVal]) -> list:
        retlist = []
        for val in vals:
            groupnr = val.value
            retlist.append(match.group(groupnr))

        return retlist


    def _convert_message(self, message, match):

        typeman = match.group(RegVal.type.value)

        if "BREAKING CHANGE:" in message and not match.group(RegVal.exclamation.value):
            boom = emojis.emojis.db.get_emoji_by_alias("boom").emoji
            mslist = message.split('\n')
            mslist[0] =  boom + mslist[0]
            message = "\n".join(mslist)
            # message = re.sub(firval, firval+"!", message)


        has_emoji = bool(emojis.get(typeman))
        has_match = typeman in self.emomap
        if has_emoji:
            pass
        elif has_match:
            emo = emojis.encode(self.emomap[typeman])
            s = r"^\b{}\b".format(typeman)
            message = re.sub(s, emo, message)
        else:
            print("Couldn't convert to emoji.")

        return message


    def get_reg(self, types=False, mid=False, scopes=False):
        pass

    def msg_dict_to_str(self, answers: dict) -> str:
        """Generate the message with the given answers."""

        # if answers["issue_close"]:
        #     answers["issue"] = "Closes {}".format(answers["issue"])
        # else:
        #     answers["issue"] = ""

        if answers["is_breaking_change"]:
            answers["breaking_change"] = "BREAKING CHANGE:"
            answers["prefix"] = answers["prefix"] + "!"

        else:
            answers["breaking_change"] = ""

        # ret = "{prefix}{scope} {subject}\n\n{breaking_change}{body} \n\n{footer}{issue}".format(
        #     **answers
        # )
        ret = "{prefix}{scope} {subject}\n\n{breaking_change}{body}".format(
            **answers
        )
        return ret



class RegValidator(Validator):
    # Ugly and weird validator, not relying on other parts of regex tool.
    # Will polish in the future, works for now.

    def __init__(self):
        self.msgtool = MsgTool()

    def find_error(self, msg):
        typereg = '([\u263a-\U0001f645]+)'
        if not re.match(f"{typereg}", msg):
            return "Insert type first."
        elif re.match(f"{typereg}\([a-z,]*$", msg):
            return "Finish scope part."
        elif re.match(f"{typereg}\([a-z,]*\)$", msg):
            return "Need space between type/scopes and subject."
        elif re.match(f"{typereg}(\([a-z,]*\))* $", msg):
            return "Subject can't be empty."
        elif not self.msgtool.check_message(msg):
            return "Message currently invalid."

    def validate(self, document: Document) -> None:
        errmsg = self.find_error(document.text)
        if errmsg:
            raise ValidationError(message=errmsg)



import os
import re

import emojis
from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit.completion import Completer, Completion, FuzzyCompleter

from wommit.utils import utils
import questionary
import pathlib

class Configurator:
    # Should never edit global config unless global_conf is True.
    def __init__(self, testing: bool=False, global_conf: bool=False):

        self.testing = testing
        self.global_conf = global_conf

        if (global_conf and not os.getenv('SHAPEIT_CONF')):
            os.environ['SHAPEIT_CONF'] = str(pathlib.Path.home())

        self.data = utils.load_data(self.global_conf)

    def _write_data(self, data):
        utils.write_data(data, self.global_conf)

    def load_data(self):
        return self.data
      
    def _write_data(self, data):
        utils.write_data(data, self.global_conf)

    def settings(self):
        questions = [
            {
                "type": "select",
                "name": "option",
                "message": "What setting do you wish to edit?",
                "choices": ["default commit method", "use global config"],
            },
            # Add part
            {
                "when": lambda x: x["option"] == "default commit method",
                "type": "select",
                "name": "which_method",
                "message": "Which method should be the default?",
                "choices": ["menu", "autocomplete"],
            },

            {
                "when": lambda x: x["option"] == "use global config",
                "type": "select",
                "name": "use_global",
                "message": "Use global config by default?",
                "choices": ['Yes', 'No'],

            },

        ]

        s = questionary.prompt(questions)

        if "which_method" in s:
            method = s['which_method']
            self.data.update({'settings':{'method':method}})
            self._write_data(self.data)

        elif "use_global" in s:
            answer = s['use_global']
            boolval = True if answer == 'Yes' else False
            data = dict(self.data, **{'use_global':boolval})
            self._write_data(data)
            print("Local settings now used!")

    # def _get_global_setting(self):


    def print_info(self):
        type_lines = [dicty["value"] for dicty in self.data["types"]]
        formatted_types = "\n".join(type_lines)
        formatted_scopes = "\n".join(self.data["scopes"])

        if utils.is_global_setting():
            state = "GLOBAL"
        else:
            state = "LOCAL"
        print_formatted_text(HTML(f"<sss>{state}</sss>\n<aaa bg='ansiwhite' fg='ansiblack'>Types:</aaa>\n{formatted_types}\n\n"
                                  f"<aaa bg='ansiwhite' fg='ansiblack'>Scopes:</aaa>\n{formatted_scopes}"))

    def edit(self):
        questions = [
            {
                "type": "select",
                "name": "option",
                "message": "What do you wish to do?",
                "choices": ["delete", "add"],
            },
            # Add part
            {
                "when": lambda x: x["option"] == "add",
                "type": "select",
                "name": "which",
                "message": "What do you wish to add?",
                "choices": ["type", "scope"],
            },
            {
                # "when": lambda x: x["which"] == "scope",
                "when": lambda x: "scope" in x.values(),
                "type": "text",
                "name": "addedscope",
                "message": "Type in the name of the scope: ",
                "filter": lambda y: re.sub(
                    r"\W+", "", y
                ),  # Removes all non alphanumeric characters.
            },
            {
                "when": lambda x: "type" in x.values(),
                "type": "text",
                "name": "addedtype",
                "message": "Type in the name of the type: ",
                "filter": lambda y: re.sub(
                    r"\W+", "", y
                ),  # Removes all non alphanumeric characters.
            },
            {
                "when": lambda x: "addedtype" in x,
                "type": "text",
                "name": "type_info",
                "message": "Explain your type:",
            },

            # {
            #     "when": lambda x: "addedtype" in x,
            #     "type": "autocomplete",
            #     "name": "type_emoji",
            #     "message": "Get an emoji for your type:",
            #     "completer": FuzzyCompleter(self.EmojiCompleter),
            #     "choices": ["test"],
            # },

            {
                "when": lambda x: "addedtype" in x,
                "type": "autocomplete",
                "name": "type_emoji",
                "message": "Get an emoji for your type:",
                # "completer": FuzzyCompleter,
                "completer": self.EmojiCompleter(),
                # "choices": [name for emoji in emojis.emojis.db.db.EMOJI_DB for name in emoji.aliases],
                "choices": ["asd"],
                # "choices": [emoji.alises for emoji in emojis.emojis.db.db.EMOJI_DB for emoji in emoji],
            },

            # Delete part
            {
                "type": "checkbox",
                "when": lambda x: x["option"] == "delete",
                "name": "delvals",
                "message": "Which options do you want to remove?",
                "choices": self._format_q(),
            },
        ]

        s = questionary.prompt(questions)
        if "delete" in s.values():
            vals_deleted = self._del_values(s["delvals"])
            print("Deleted: ")
            print(", ".join(vals_deleted))
        # elif s['option'] == 'add':
        else:
            if "addedscope" in s:
                addvar = s["addedscope"]
                self.data["scopes"].append(addvar)

            elif "addedtype" in s:
                # Add type
                addvar = s["addedtype"]
                addtypeinfo = s["type_info"]
                dicadd = {"value": addvar, "name": f"{addvar}: {addtypeinfo}"}
                self.data["types"].append(dicadd)

                # Add emoji for type
                addemoji = s["type_emoji"]
                self.data["emoji_map"][addvar] = addemoji

            if self.testing:
                print("Test finished, won't add to data.")
            else:
                self._write_data(self.data)
                which = s["which"]
                print(f"Added '{addvar}' as an available {which}.")

    def _format_q(self):

        listman = list()
        listman.append(questionary.Separator("Types:"))
        types = [a["value"] for a in self.data["types"]]
        type_dict_list = [{"name": a, "value": "types-" + a} for a in types]
        listman.extend(type_dict_list)

        listman.append(questionary.Separator("Scopes:"))
        scopes = self.data["scopes"]
        scope_dict_list = [{"name": a, "value": "scopes-" + a} for a in scopes]
        listman.extend(scope_dict_list)

        # print(listman)

        return listman

    def _del_values(self, delvalues) -> list:
        """
        Delete values from data file.
        :param delvalues: values to be deleted
        :return: deleted all files that got deleted
        """
        # print(delvalues)
        actualdelvals = list()
        for delval in delvalues:
            splitboy = delval.split("-")
            category, value = splitboy[0], splitboy[1]
            if category == "types":
                # delete list dict element
                self.data[category][:] = [
                    d for d in self.data[category] if d.get("value") != value
                ]

            else:
                self.data[category] = [d for d in self.data[category] if d != value]

            actualdelvals.append(value)

        if not self.testing:
            self._write_data(self.data)
        else:
            print("Test completed, won't save changes.")
        return actualdelvals

    def overwrite_with_global(self):
        """Overrides local data with global data."""
        newconf = Configurator(global_conf=True)
        globaldata = newconf.load_data()
        self._write_data(globaldata)
        return globaldata


    class EmojiCompleter(Completer):

        def __init__(self):
            self.all_emojis = emojis.emojis.db.db.EMOJI_DB

        def get_completions(self, document, complete_event):
            word = document.get_word_before_cursor()

            for emo in self.all_emojis:
                for alias in emo.aliases:
                    if str(alias).startswith(word):
                        yield Completion(text=emo.emoji,display_meta=alias, start_position=-len(word))



import re
import subprocess
import time
from typing import Iterable

import emojis
from prompt_toolkit import HTML
from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.styles import Style

from wommit.utils.exceptions import NoRepoAccessError
from wommit.apihandler import ApiHandler
import wommit.utils.utils as utils
from wommit.utils.utils import load_data, get_format_string

from prompt_toolkit.completion import Completer, Completion, CompleteEvent
import concurrent.futures
from wommit.msgtool import MsgTool
import logging



class WommitCompleter(Completer):
    # #
    def __init__(self, data=None):
        if data:
            self.data = data
        else:
            self.data = load_data()
        self.regtool = MsgTool(self.data)

        logging.basicConfig(filename='myapp.log', level=logging.INFO)

        self.handler = ApiHandler()
        self.issuecomp = IssueCompleter(self.handler)
        self.commitcomp = CommitCompleter(self.handler)

    def form_meta(self, data):
        pass

    def get_completions(self, document, complete_event):

        before = document.text_before_cursor
        word = document.get_word_before_cursor()
        charman = document.char_before_cursor

        # Give all types at start
        # if document.line_count == 1:
        if document.cursor_position_row == 0:

            typematch = re.match("(\w+)$", before)
            scopereg = "^([\u263a-\U0001f645]+)\({1}[a-z,]*"
            scopematch = re.match(scopereg+"$", before)
            endscopematch = re.match(f"{scopereg}\)$", before)
            # scopematch = re.match("^([\u263a-\U0001f645]+)\({1}[a-z,]*$", before)

            # THESE HAVE TO BE IN REVERSE PRIORITY ORDER


            # # Scope end
            # if re.match(self.type_reg + self.scope_start_reg + self.scope_end_reg + "$", document.text_before_cursor):
            if endscopematch:
                yield Completion(") ", start_position=-1)

            # Scopes
            elif scopematch:
                for scope in self.data["scopes"]:
                    if charman == "(":  # Special case at start, yield all
                        start_position = 0
                    elif scope not in before:
                        if charman == ",":
                            start_position = 0
                        elif scope.startswith(word):
                            start_position = -len(word)
                        else:
                            continue
                    else:
                        continue

                    yield Completion(
                        text=scope, display=scope, start_position=start_position
                    )

            #Types
            elif typematch or document.cursor_position == 0:
                for type in self.data["types"]:
                    val = type["value"]
                    if val.startswith(word):
                        if val in self.data['emoji_map']:
                            x = self.data['emoji_map'][val]
                            val_emoji = emojis.encode(x)
                            yield Completion(text=val_emoji, display_meta=val, start_position=-len(word))
                            # yield Completion(text=val, display_meta=val_emoji, start_position=-len(word))


        # Closing
        else:


            comps = list(self.issuecomp.get_completions(document, complete_event))
            commit_comps = list(self.commitcomp.get_completions(document, complete_event))

            if comps:
                yield from comps

            elif commit_comps:
                yield from commit_comps


            elif (word.lower()).startswith("clo"):
                yield Completion(
                    text="Closes ", display="Closes", start_position=-len(word)
                )

            # elif (word.lower()).startswith("bre") and before=="":
            elif document.current_line_before_cursor.lower() == "bre":
                yield Completion("BREAKING CHANGE: ", start_position=-len(word))



class CommitCompleter(Completer):

    def __init__(self, handler: ApiHandler):
        self.commits = handler.get_commits()

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:

        # before = document.text_before_cursor.lower()
        currline = document.current_line_before_cursor
        word = document.get_word_before_cursor().lower()
        charman = document.char_before_cursor

        for commit in self.commits:
            prog = re.compile(r"\(([a-zA-Z0-9]*)\)?", re.MULTILINE)
            regcheck = prog.finditer(currline)
            # titlecheck = commit.title.startswith(word)


            if regcheck:
                for match in regcheck:
                    if ')' not in match.group(0):

                        logging.info(f"content: {match.start()}, {match.end()} " +str(match.group(0)))
                        shacheck = str(commit.sha).startswith(word)
                        if charman == '(' or shacheck:
                            # start_position = 0 if charcheck else -len(word)
                            start_position = -len(word) if charman == '(' else -len(word)-1
                            completion = Completion(
                                text='('+commit.sha+')',display=commit.sha, display_meta=commit.title, start_position=start_position)

                            yield completion



class IssueCompleter(Completer):

    def __init__(self, handler: ApiHandler):
        self.issues = []
        self.future = concurrent.futures.ThreadPoolExecutor().submit(handler.get_issues)
        self.done = False


    def get_completions(self, document, complete_event):

        before = document.text_before_cursor.lower()
        word = document.get_word_before_cursor().lower()

        wordmatch = re.match(r'[\s\S]*#([a-z]+$)', before)
        digmatch = re.match(r'[\s\S]*#([\d]*$)', before)
        if digmatch or wordmatch:
            logging.info(f"digmatch: {bool(digmatch)}!r")
            logging.info(f"wordmatch: {bool(wordmatch)}!r")
            # If thread is not complete
            if not self.done:
                self.fetch_issues()

            # If there are issues to go through
            if self.issues:
                    # There  is a # or a reg match
                for type, nr, title in self.issues:



                    if wordmatch:
                        firval = wordmatch.group(1).lower()
                        start_position = -len(word) - 1
                    else:
                        firval = digmatch.group(1)
                        start_position = -len(word)

                    # if letters typed
                    logging.info(f"title: {title}")
                    logging.info(f"firval word: {firval}")
                    ret_word_b = wordmatch and (title.lower().startswith(firval))
                    ret_dig_b = digmatch and (not firval or str(nr).startswith(firval))

                    if ret_word_b or ret_dig_b:

                        # Styling
                        if type == "PR":
                            style = "fg: #f7f719"
                            selected_style = "bg:#f7f719 fg:ansiblack"
                            info_bg = "#f7f719"


                        else:
                            style = "fg:#0ee6e6"
                            selected_style = "bg:#0ee6e6 fg:ansiwhite"
                            info_bg = "#0ee6e6"

                        info_html = HTML(f"<ttt fg='ansiwhite' bg='#3b8a40'>{title} </ttt><aaa bg='{info_bg}'>{type}</aaa> ")

                        completion = Completion(
                            nr, display_meta=to_formatted_text(info_html), start_position=start_position, style=style,
                            selected_style=selected_style)

                        yield completion
            else:
                return False

    def fetch_issues(self):
        if self.future.done():
            try:
                self.issues = self.future.result()
            except NoRepoAccessError as e:
                logging.info("Could not fetch issues.")
                logging.info(e.message)

            self.done = True

def get_prompt_info(show_staged: bool=False):

    text = f"""
<one>Write your commit.</one> <two>Required format</two>:
<format bg='lightgray'>{get_format_string(as_html=True)}</format>

<end>Press Tab to request completion. \nPress Esc+Enter to finish, or Ctrl+C to abort.</end>

> """


    if show_staged:
        runargs = utils.get_subprocess_runargs()
        call = subprocess.run("git diff --name-only --cached".split(" "), **runargs)
        out = call.stdout
        text = f"\nCommited files:\n<files>{out}</files>{text}"

    style = Style.from_dict(
        {
            "one": "#57D900",
            "two": "#ff0066 underline",
            "format": "black",
            "end": "ansicyan",
            "files": "#808080"
        }
    )

    # message = [('class.one':'Write your commit.'),'']
    return text, style



import time

from prompt_toolkit import PromptSession, HTML
from prompt_toolkit.application import get_app
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.validation import Validator

from wommit.apihandler import ApiHandler
from wommit.autocomplete import WommitCompleter, get_prompt_info, IssueCompleter
from wommit.utils.exceptions import NoAnswersError, NoStandardsError
from wommit.msgtool import MsgTool
from wommit.utils.utils import load_data, emojize_types
from questionary import prompt

import wommit.msgtool


# def parse_subject(text):
#     if isinstance(text, str):
#         text = text.strip(".").strip()
#
#     return required_validator(text, msg="Subject is required.")


def parse_scope(scopes):
    if not scopes:
        return ""

    f_scope = ",".join(scopes)
    return f"({f_scope})"


class Questioneer:
    def __init__(self, data=None, show_staged:bool = False):
        if not data:
            self.data = load_data()
        else:
            self.data = data

        self.show_staged = show_staged

        self.msgtool = MsgTool(self.data)

    def autocomp(self):
        """Shape commit message using autocompletion."""
        prompttext, style = get_prompt_info(self.show_staged) # Give better name for prompt_stuff. E.g compose_info, compose_prompt
        completer = WommitCompleter(self.data)
        # The creation of the session is what takes the majority of the startup time.
        session = PromptSession(completer=completer)

        text = session.prompt(
            HTML(prompttext),
            style=style,
            complete_while_typing=True,
            multiline=True,
            # pre_run=session.default_buffer.start_completion(),
            # bottom_toolbar=self.bottom_toolbar(session.history.get_strings()),
            # validator=Validator.from_callable(
            #     self.regtool.check_message, error_message="Does not yet meet standards."
            # ),
            validator=wommit.msgtool.RegValidator(),
            key_bindings=self._get_kb(),
            auto_suggest=AutoSuggestFromHistory()
        )
        # return self.do_check(text)
        # return text
        return self.msgtool.check_and_convert(text)

    def select(self):
        """Shape commit message using menus."""
        answers = prompt(self.questions())

        # If the user interruptes
        if not answers:
            raise NoAnswersError

        # message = self.regtool.format_message(answers)

        f = self.msgtool.check_and_convert(answers)
        if not f:
            raise NoStandardsError(
                f"Commit message does not meet standards.\n{f}"
            )

        return f

    def questions(self) -> list:
        """Questions regarding the commit message."""

        repr_data = emojize_types(self.data)

        issuecomp = IssueCompleter(ApiHandler())

        questions = [
            # Parses feature
            {
                "type": "list",
                "name": "prefix",
                "message": "What type of commit is this:",
                "choices": repr_data["types"],  # Gets listed features in info json.
            },
            # Parses scope
            {
                "type": "checkbox",
                "name": "scope",
                "message": ("Scope. Specify place of commit change: \n"),
                "choices": repr_data["scopes"],  # Gets listed scopes in info json.
                "filter": parse_scope,
            },
            # Parses commit description
            {
                "type": "input",
                "name": "subject",
                "message": "The subject of this commit:\n",
                # 'filter': parse_subject,
                "validate": Validator.from_callable(
                    (lambda x: len(x) > 0), error_message="Subject can't be empty."
                ),
            },
            # Body of commit
            {
                "type": "autocomplete",
                "name": "body",
                "completer": issuecomp,
                "choices": ['asd'],
                "message": (
                    "Body. Motivation for the change and contrast this "
                    "with previous behavior:\n"
                ),
            },
            # # Optional later addition of breaking changes.
            {
                "type": "confirm",
                "message": "Is this a BREAKING CHANGE? Correlates with MAJOR in SemVer",
                "name": "is_breaking_change",
                "default": False,
            },
            # # Optional footers
            # {
            #     "type": "input",
            #     "name": "footer",
            #     "message": (
            #         "Footer. Information about Breaking Changes and "
            #         "reference issues that this commit closes:\n"
            #     ),
            # },
            # Issue
            # {
            #     "type": "confirm",
            #     "name": "issue_close",
            #     "message": "Does this close an issue?",
            #     "default": False,
            # },
            # {
            # "type": "autocomplete",
            #         "name": "issue",
            # "message": "Which one?",
            # "completer": issuecomp,
            # "choices": ['asd'],
            # "when": lambda x: x["issue_close"],
            # "validate": lambda val: re.compile("#\d").match(val) is not None,
            # # 'filter': parse_issue
            # },
        ]
        return questions

    def _get_kb(self):
        kb = KeyBindings()

        @Condition
        def is_line1():
            return get_app().current_buffer.document.line_count == 1

        @kb.add('c-m', filter=is_line1)
        def _(event):
            " Initialize autocompletion, or select the next completion. "
            buff = event.app.current_buffer
            buff.newline()
            buff.newline()

        @Condition
        def one_comp():
            buff = get_app().current_buffer
            if buff.complete_state:
                comps = buff.complete_state.completions
                return len(comps) == 1 and '(' not in buff.document.text_before_cursor
            else:
                return False

        @kb.add('space', filter=one_comp)
        def _(event):
            " Initialize autocompletion, or select the next completion. "
            buff = event.app.current_buffer
            # comp = buff.go_to_completion(0)
            comp = buff.complete_state.completions[0]
            buff.apply_completion(comp)
            buff.insert_text(" ")
            # word = buff.document.get_word_before_cursor()
            # buff.insert_text((comp.text + " "), overwrite=True, move_cursor=-(len(word)))


        return kb




import re

import click
#import questionary
import questionary
from click import ClickException
from click.testing import CliRunner
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.styles import Style

from wommit.apihandler import ApiHandler
from wommit.commitman import Questioneer
#import shutil
import subprocess

from wommit.msgtool import MsgTool
from wommit.utils.changelog_gen import ChangelogGenerator, LogVersion
from wommit.utils.exceptions import NoAnswersError, NoStandardsError, RepoException, NoGitRepo
from wommit.utils.utils import get_format_string
import wommit.utils.utils as utils
from wommit.configman import Configurator
from prompt_toolkit import print_formatted_text, HTML
import hashlib


@click.group()
def cli():
    """A package for intuitively commiting files with Git."""
    pass

@cli.command()
@click.option("-id", type=str, help="Check a commit message with the specified ID.")
@click.option("-ids", nargs=2, type=str, default="", help="Check all commit messages between two IDs (newest ID first).")
@click.option("-m", type=str, help="Check if the given string passes the check.")
@click.option("-l", "--latest", is_flag=True, help="Check all local commits that have not been pushed.")
def check(id, ids, m, latest):
    """Checks if previous commit message meets the Conventional Commits format."""

    try:

        # pop = subprocess.Popen(text=True, check=True, capture_output=True, encoding='utf-8')
        runargs = utils.get_subprocess_runargs()
        gitcmd = "git log -1 --pretty=%B".split()
        check_msgs = []

        if id:
            gitcmd.append(id)
            msg = subprocess.run(gitcmd, **runargs).stdout
            check_msgs.append(msg)
        elif m:
            check_msgs.append(m)
        elif latest or ids: # Format this piece of code better.
            if ids:
                hashgets = f"git log {ids[1]}..{ids[0]} --pretty=%H".split(' ')
            else:

                branchname = subprocess.run('git branch --show-current'.split(), **runargs).stdout.strip()
                hashgets = f"git log origin/{branchname}..{branchname} --pretty=%H".split(' ')
            callhashes = subprocess.run(hashgets, **runargs).stdout.strip()
            hashes = [x for x in callhashes.split("\n") if x]
            if not hashes:
                print("No new commits to check.")
                return
            for hash in reversed(hashes):
                msg = subprocess.run(gitcmd + [hash], **runargs).stdout
                check_msgs.append(msg)

        else: # Default check on latest commit
            msg = subprocess.run(gitcmd, **runargs).stdout
            check_msgs.append(msg)

        # Check all git messages.

        text = FormattedText([('', '\n'),('bg:#D924EF' , 'SHAPEIT CHECK'),('','\n')])

        print_formatted_text(text)

        for msg in check_msgs:
            _check_and_print(msg)

    except subprocess.CalledProcessError as e:

        # r = re.find("\bfatal: ambiguous argument '[\s\S]+': unknown revision or path not in the working tree.\b")
        # if "fatal: ambiguous argument 'asd': unknown revision or path not in the working tree" in e.stderr:
        if "fatal: ambiguous argument " in e.stderr:
            print(e.stderr)
            raise click.BadParameter("Found no commits with this hash", param=id, param_hint="-i")

        elif "fatal: not a git repository (or any of the parent directories): .git" in e.stderr:
            raise click.UsageError("This is not a git repository. Make it a git repository with git init")

        else:
            raise e

def _check_and_print(msg):

    tool = MsgTool()

    # msg = tool.check_and_convert(msg)
    cc_msg = tool.check_and_convert(msg)
    if cc_msg:
        passtup = ("ansigreen" , "Your message passes! :)")
    else:
        passtup = ("ansired", "Your message doesn't pass! :(")

    text = FormattedText([

        # ('bg:#D924EF' , 'SHAPEIT CHECK'),
        # ('','\n'),
        # ('#4DF5F7', 'Your commit:\n\n'),
        ('' , msg),
        ('', '\n'),
        passtup,
        ('', '\n\n---\n'),
    ])

    style = Style.from_dict({
        # 'msg':'bg:#A6ACAC fg:#ansiblack',
        'msg': 'fg:#ansiwhite',
    })

    print_formatted_text(text, style=style)

@cli.command()
def format():
    """Prints the Wommit commit format."""
    print("Format:\n" + get_format_string() + "\n")

@cli.group()
@click.option('-t', is_flag=True, help="Test the functionality.")
@click.option('-g', is_flag=True, help="Global config.")
@click.pass_context
def configure(ctx, t, g):
    """Configure available types and scopes."""
    ctx.ensure_object(dict)
    ctx.obj['confman'] = Configurator(t, g)

@configure.command()
@click.pass_context
def e(ctx):
    """Edit current types and scopes."""
    try:
        ctx.obj['confman'].edit()
    except KeyError as e: # Process was cancelled and returned empty values.
        exit()

@configure.command()
@click.pass_context
def p(ctx):
    """Prints all types and scopes."""
    ctx.obj['confman'].print_info()

@configure.command()
@click.pass_context
def s(ctx):
    """Edit settings."""
    try:
        ctx.obj['confman'].settings()
    except KeyError as e: # Process was cancelled and returned empty values.
        exit()

@configure.command()
@click.pass_context
def overwrite(ctx):
    """Overwrites local data with global."""
    val = questionary.confirm("Are you sure").ask()
    if val:
        globaldata = ctx.obj['confman'].overwrite_with_global()
        print("Global data is now local!")

@cli.command()
@click.option('-e', '--expert', is_flag=True, help="Expert mode, autocompletion. Overrides default.")
@click.option('-m', '--menu', is_flag=True, help="Menu mode. Overrides default.")
@click.option('--manual', type=str, help="Custom message input, converts feat names to emojis.")
@click.option('-t', '--test','--dry', is_flag=True, help="Manually test the functionality.")
@click.option('-g', is_flag=True, help="Use global settings.")
def c(expert, menu, manual, test, g):
    """Commit your staged files using an intuitive menu."""
    try:
        if not test and not utils.has_staged_files():
            print("Your repo has no files staged for commiting.")
            return

        conf = Configurator(testing=test, global_conf=g)
        data = conf.load_data()
        q = Questioneer(data=data, show_staged=True)
        default_mode = data['settings']['method']
        if manual:
            message = MsgTool().check_and_convert(manual)

        elif expert:
            message = q.autocomp()
        elif menu:
            message = q.select()
        elif default_mode == "autocomplete":
            message = q.autocomp()
        elif default_mode == 'menu':
            message = q.select()



    except NoAnswersError:# The user interrupted the prompt.
        raise click.Abort
    except NoStandardsError as err: # If the message didn't follow the standards.
        raise click.UsageError("Message did not conform to the commit standards.")
    except (RepoException, NoGitRepo) as err:
        raise click.UsageError("You are not inside a git repo.")

    # Er allerede sjekket. Trenger bare å printe.
    _check_and_print(message)
    # print("Commit message passed.")

    if test:
        print(f"\nTest completed.\n")

    else:

        try: # Attempt to commit
            subprocess.run(['git', 'commit', '-m', message], text=True, check=True)
        except subprocess.CalledProcessError as expert:
            print(expert.stderr)
        else:
            print_formatted_text(HTML("<aaa bg='ansigreen'>Commited!</aaa>"))

@cli.command()
@click.option('-b','--basic',is_flag=True, help="Prints an easily readable changelog.")
@click.option('-r','--release',is_flag=True, help="Prints a changelog designed for Github releases.")
@click.option('-e', '--edit', is_flag=True, help="Edit a release tag with the generated changelog.")
@click.option('-post', '--post_release', is_flag=True, help="Checks between two latest instead of latest to now.")
def changelog(basic, release, edit, post_release):
    """Generates a changelog from all wommit commits."""

    if basic:
        gen = ChangelogGenerator(LogVersion.BASIC)
    elif release:
        gen = ChangelogGenerator(LogVersion.RELEASE)
    else:
        gen = ChangelogGenerator(LogVersion.BASIC)

    changelog = gen.generate_changelog(pre_release=(not post_release))

    if edit:
        print("going in")
        ApiHandler().edit_release(changelog)

    try:
        print(changelog)
    except:
        return changelog


import os
import string
import random
import subprocess
from collections import namedtuple

import keyring
import questionary
import requests
from requests.models import PreparedRequest

# from httpx import Request # Possible future import to replace requests

from requests.adapters import HTTPAdapter
from urllib3 import Retry
import webbrowser

import logging
from multiprocessing import Queue
from webtest.http import StopableWSGIServer
import falcon

from wommit.utils import utils
from wommit.utils.utils import get_repo_info
from wommit.utils import utils
from wommit.utils import decos


# Used to store PR and issue data.
Issue = namedtuple("Issue", ["type", "number", "title"])
Commit = namedtuple("Commit", ["sha", "title"])


class ApiHandler:

    def __init__(self):
        self.keyring_name = "git:https://github.com"
        self.keyring_username = "PersonalAccessToken"
        self.graphql_api_url = "https://api.github.com/graphql"
        self.token = self.get_token()

    def get_commits(self):

        variables = get_repo_info()

        api_url = "https://api.github.com/graphql"

        runargs = utils.get_subprocess_runargs()
        branchname = subprocess.run('git branch --show-current'.split(), **runargs).stdout.strip()

        query = """query($owner:String!, $name:String!) {
      repository(owner: $owner, name: $name) {
        refs(refPrefix: "refs/heads/", orderBy: {direction: DESC, field: TAG_COMMIT_DATE}, first: 100) {
          totalCount
          edges {
            node {
              name
              target {
                ...on Commit {
                  history(first:10){
                  edges{
                  node{
                  messageHeadline
                  id
                  abbreviatedOid
                  }
                  }
                    totalCount
                  }
                  
                }
              }
            }
          }
          pageInfo {
            endCursor
            hasNextPage
          }
        }
      }
    }"""

        r = self._graphql_func(query, variables)

        r_data = r.json()

        commits = []

        for node in r_data['data']['repository']['refs']['edges']:
            currbranch = node['node']['name']
            if currbranch == branchname:
                for commit in node['node']['target']['history']['edges']:
                    commit_msg = commit['node']['messageHeadline']
                    id = commit['node']['abbreviatedOid']
                    # print(f"--> {commit_msg} , {id}")
                    commits.append(Commit(id, commit_msg))

        return commits


    def get_issues(self) -> dict:

        api_url = "https://api.github.com/graphql"

        variables = get_repo_info()
        if not variables:
            return {}

        #token = self.get_token()
        # token = "abc"

        query = """
        query ($owner : String!, $name : String!){
          repository(owner: $owner, name: $name) {
            issues(last:10, states:OPEN) {
              edges {
                node {
                  title
                  number
                }
              }
            }
            pullRequests(last: 10, states: OPEN){
              edges {
                node {
                  title
                  number
                }
              }
            }
          }
        }
        """

        r = self._graphql_func(query, variables)


        r_data = r.json()

        if str(r.status_code) == "401":
            print("Local access token is incorrect for user credentials.")
            return

        elif 'errors' in r_data:
            print("Token does not have access to Github repo, or it doesn't exist.")
            logging.info(r_data['errors'])

            # r = r_func(self.oauth_flow())
            #
            # r_data = r.json()

            # if questionary.confirm(
            #     "Do you want to try a manual access token?", default=True
            # ).ask():
            #     self.get_token_manually()
            # else:
            #     return
            return

            # r = r_func(self.get_token_manually())

            # a31166d848e219933c08a0c3d2cd69cefe6f9a5f
            # raise NoRepoAccessError



        issue_nodes = r_data["data"]["repository"]["issues"]["edges"]
        pr_nodes = r_data["data"]["repository"]["pullRequests"]["edges"]



        def make_retvals(valname, allvals) -> list:

            return [
                Issue(valname, "#" + str(inner_d["number"]), inner_d["title"])
                for node in allvals
                for inner_d in node.values()
            ]

        complist = [*make_retvals("issue", issue_nodes), *make_retvals("PR", pr_nodes)]
        retvals = sorted(complist, key=lambda x: x.number)

        return retvals

    def get_token(self):
        """ Accesses local Windows Certificates to return a Github access token.
        :return: token str
        """

        # NOTE: this code currently returns a access token with incorrect scopes, and needs manual verification from
        # bkkp to access bkkp/valuable, which CLI doesn't. Working on resolving.
        #

        token = self.get_local_token()
        if token:
            return token
        else:
            print(f'Could not find a git access token under {self.keyring_name} {self.keyring_username}.')
            # token = self.get_token_manually()
            token = self.oauth_flow()


        return token

    def get_local_token(self):
        # known_names = ['Personal Access Token', 'PersonalAccessToken']
        # for n in known_names:
        try:
            k_key = keyring.get_credential(self.keyring_name, self.keyring_username).password
            return k_key
        # token = keyring.get_password("git:https://github.com", n)
        except Exception:
            print("checking environment variable")
            e_key = os.getenv("GITHUB_TOKEN")
            return e_key

    def get_token_manually(self):
        return input("Get token from Github and paste here:")

    def oauth_flow(self):

        # https://developer.github.com/apps/building-oauth-apps/authorizing-oauth-apps/
        # 1
        link = "https://github.com/login/oauth/authorize"
        client_id = "c60cc3da378a890cab46"
        # Secret added by justification of Github devs: https://github.com/cli/cli/issues/1245#issuecomment-652907737
        client_secret = "a96aac7b2cede42ab754b21414f89c22add8db91"

        port = 8080
        callback_name = "/callback"
        #
        redirect_uri = "http://localhost:{}{}".format(port, callback_name)
        token_url = "https://github.com/login/oauth/access_token"
        # Gen random state string
        letters = string.ascii_lowercase
        state = "".join(random.choice(letters) for i in range(20))
        # scopes
        scope = "read:org repo gist"

        # Authorize user
        authorize_params = {
            "client_id": client_id,
            "redirect_uri": redirect_uri,
            "scope": scope,
            "state": state,
        }
        req = PreparedRequest()
        req.prepare_url(link, authorize_params)
        webbrowser.open(req.url)

        ret_params = HttpHandler(port, callback_name).run()

        code = ret_params["code"]
        print("got the code " + code)

        # 2
        post_params = {
            "client_id": client_id,
            "client_secret": client_secret,
            "code": code,
            "state": state,
        }
        sesh = requests.Session()
        retry = Retry(connect=3, backoff_factor=0.5)
        adapter = HTTPAdapter(max_retries=retry)
        sesh.mount("http://", adapter)
        sesh.mount("https://", adapter)
        r = sesh.post(token_url, params=post_params, verify=False)
        # print(r.text)
        # make this work sometime
        # print(r.json())
        # print(r.text)
        token = self.find_between(r.text, "=", "&")
        # print(token)
        # print("new token is: " + token)

        # Sets the environment variable for future use. YOLO
        keyring.set_password(self.keyring_name, self.keyring_username, token)

        # 3
        return token

    def find_between(self, s, first, last):
        """This is absolute shite, requests.json() just isnt working"""
        try:
            start = s.index(first) + len(first)
            end = s.index(last, start)
            return s[start:end]
        except ValueError:
            return ""

    def _get_issues_res(self) -> dict:
        """Deprecated REST API version."""

        api_url = "https://api.github.com/repos/bkkp/valuable/issues"
        token = self.get_token()
        r = requests.get(
            api_url,
            headers={"Authorization": "token {}".format(token)},
            params={"state": "open"},
        ).json()
        issue_numbers = ["#" + str(a["number"]) for a in r]

        return issue_numbers

    def edit_release(self, changelog):
        """Edit latest release."""

        infodict = utils.get_repo_info()
        url_str = "https://api.github.com/repos/{owner}/{name}/releases/latest".format(**infodict)
        print(url_str)
        latest_release_url = requests.get(url_str,
                                        headers={"Authorization": f"token {self.token}"})
        lr_data = latest_release_url.json()
        # return
        print(lr_data)

        infodict['release_id'] = lr_data['id']
        url = "https://api.github.com/repos/{owner}/{name}/releases/{release_id}".format(**infodict)

        data = {'name': lr_data['tag_name'], 'target_commitish': lr_data['target_commitish'], 'tag': lr_data['tag_name'],
                'draft': False, 'prerelase': False, 'body': changelog}

        r = requests.patch(url, json=data, headers={"Authorization": f"token {self.token}"})

        r_data = r.json()

        # print(r_data)

    def get_releases(self, amount: int):



        variables = get_repo_info()
        variables['amount_rel'] = amount

        query = """
                query($owner : String!, $name : String!, $amount_rel : Int!){
                    repository(owner: $owner, name: $name){
                        releases(first: $amount_rel, orderBy: {direction: DESC, field: CREATED_AT},){
                            edges{
                                node{
                                    name
                                    tagName
                                    description
                                }
                            }
                        }

                    }
                }
                    """

        data = self._graphql_func(query, variables).json()
        #print(data)
        return data['data']['repository']['releases']['edges']

    def _graphql_func(self, query, variables):
        token=self.token
        return requests.post(
            self.graphql_api_url,
            json={"query": query, "variables": variables},
            headers={"Authorization": "token {}".format(token)},
        )

class HttpHandler:
    def __init__(self, port, callback_name):
        self.que = Queue()
        self.port = port
        self.callback_name = callback_name

    def run(self):
        self.start_serve()
        return self.que.get()

    class Resource(object):
        def __init__(self, classman):
            self.classman = classman

        def on_get(self, req, resp):
            resp.body = "You can now close this window."
            resp.status = falcon.HTTP_200
            # print("getting me")
            # logging.info("we got a get")
            self.classman.close(req.params)

    def start_serve(self):
        api = application = falcon.API()
        self.spa = self.Resource(self)
        api.add_route(self.callback_name, self.spa)

        self.s = StopableWSGIServer(api, host="localhost", port=self.port)
        self.s.run()

        # Possible future threading/process solution.
        #
        # x = threading.Thread(target=self.s.run, daemon=True)
        # x.run()
        # s.create(api, host='localhost', port='8080')
        # serve(api, host='localhost', port='8080')

    def close(self, params):
        print("trying to close")
        self.que.put(params)
        self.s.shutdown()


if __name__ == "__main__":
    # print(IssueHandler().get_local_token())

    # print(ApiHandler().get_issues())
    print(ApiHandler().get_commits())

    # decos.trace(IssueHandler().get_issues())

    # print(IssueHandler().get_issues())
    # import os
    # print(os.environ['GITHUB_TOKEN'])
    # print(os.environ['GITHUB'])
    # print(IssueHandler().get_token())
    # print(IssueHandler().oauth_flow())



# General utility functions
import json
import os

import emojis
import re
import pathlib
import html
from subprocess import run, CalledProcessError

from wommit.utils.exceptions import NoGitRepo

pack_dir_path = pathlib.Path(os.path.dirname(os.path.realpath(__file__)))
infopath = pack_dir_path.joinpath('template_info.json')
default_global = pathlib.Path.home().joinpath('SHAPEIT_CONF.json')
global_infopath =os.getenv('SHAPEIT_CONF', default_global)

cwd = pathlib.Path(os.getcwd())

def has_staged_files() -> bool:
    call = run(["git", "diff", "--staged", "--quiet"], capture_output=True)
    code = call.returncode

    if code == 1:
        return True
    elif code == 0:
        return False
    elif code == 129:
        raise NoGitRepo
    else:
        # Raise exception
        call.check_returncode()


def get_repo_info() -> (str, str):
    """Returns a dict with owner and name of local repo in key-value pairs."""
    gitargs = ["git", "config", "--get", "remote.origin.url"]
    repo_url = run(gitargs, check=True, capture_output=True, text=True, encoding="utf-8").stdout
    if not repo_url:
        print("Repo does not yet have a remote")
        return

    splitted = repo_url.split("/")
    x = list(map(lambda s: re.sub(r'(\.git|\n)$', '', s), splitted))
    return {'owner': x[3], 'name': x[4]}


def add_emoji_maps():
    data = load_data()
    listman = [i['value'] for i in data['types']]
    emoji_dict = {val: '' for val in listman}
    data['emoji_map'] = emoji_dict
    write_data(data)

def is_global_setting():
    with open(infopath, "r") as f:
        content = json.load(f)
    return content["use_global"]

def load_data(g=False):

    if g:
        with open(global_infopath, "r") as f:
            content = json.load(f)
    else:

        with open(infopath, "r") as f:
            content = json.load(f)

        if is_global_setting() or g:
            if global_infopath.is_file():
                with open(global_infopath, "r") as f:
                    content = json.load(f)
            else:
                write_data(content, g)

    return content

def write_data(data : dict, g=False):
    if g:
        path = global_infopath
    else:
        path = infopath
    with open(path, "w") as f:
        json.dump(data, f, indent=4)


def emojize_types(data, name='name'):
    ret_data = dict(data)
    for typedict in ret_data['types']:
        currval = typedict['value']
        try:
            relemoji = emojis.encode(data['emoji_map'][currval])
            typedict[name] = "{} {}".format(relemoji, typedict[name])
        except KeyError as e:
            print(e)
            continue

    return ret_data

def get_subprocess_runargs():
    return {'text': True, 'check': True, 'capture_output': True, 'encoding': 'utf-8'}


# def check_message(message: str, types: list=[]) -> bool:
#     """
#     Checks if the commit message conforms to
#     the conventional commit standard.
#     :param message: commit message to check
#     :param types: mandatory types for repo. if empty, all types are allowed.
#     :return: True if the message is valid, False otherwise.
#     """
#     if types:
#         typestr = "|".join(types)
#     else:
#         typestr = ".+" # allow anything as type
#         #raise AllowedTypesError
#     reg_s = r"^({typestr})(!?)(\({{1}}[a-z,]+\){{1}})?: ([^ ]+)".format(typestr=typestr) # Double { used to escape.
#     return bool(re.match(reg_s, message))

# def check_message(message: str, types: list=[]) -> bool:
#     """
#     Checks if the commit message conforms to
#     the conventional commit standard.
#     :param message: commit message to check
#     :param types: mandatory types for repo. if empty, all types are allowed.
#     :return: True if the message is valid, False otherwise.
#     """
#     if types:
#         typestr = "|".join(types)
#     else:
#         typestr = ".+" # allow anything as type
#         #raise AllowedTypesError
#     reg_s = r"^({typestr})(!?)(\({{1}}[a-z,]+\){{1}})? ([^ ]+)".format(typestr=typestr) # Double { used to escape.
#     return bool(re.match(reg_s, message))

def convert_type_to_emoji(type:str):
    pass


def get_format_string(as_html=False):
    format = """
&lt;type&gt;[optional scope] &lt;description&gt;
[optional body]
[optional footer]"""
    if as_html:
        return format
    else:
        return str(html.unescape(format))




# Add variant of this. Currently regex is stored in two spots.
#
# def get_regex(as_string=True):
#     type_reg = '^\w+'
#     mid_reg = '\({1}[a-z,]*'
#     scope_reg = '\){1}'
#     end_reg = ':'
#
#     if as_string:
#         return f'{type_reg}{mid_reg}{scope_reg}'



# --- LOGGING UTIL ---
import functools
import logging
import time

logging.basicConfig(filename='myapp.log', level=logging.INFO)
default_logger = logging.getLogger()


def timer(logger=default_logger, level="INFO"):
    """ Credits:
    https://realpython.com/primer-on-python-decorators/#decorators-with-arguments
    """

    def decorator_timer(func):
        """ Print the runtime of the decorated function"""

        @functools.wraps(func)
        def wrapper_timer(*args, **kwargs):
            lvl = logging.getLevelName(level)
            if not isinstance(lvl, int):
                lvl = logging.getLevelName("INFO")
                logging.warning("Unrecognised logging level!")
            logger.log(level=lvl, msg=f"Calling {func.__name__}")
            start_time = time.perf_counter()
            value = func(*args, **kwargs)
            end_time = time.perf_counter()
            run_time = end_time - start_time
            logger.log(
                level=lvl, msg=f"Finished {func.__name__!r} in {run_time:.4f} secs"
            )
            return value

        return wrapper_timer

    return decorator_timer


def trace(logger=default_logger, timeit=True, level="INFO"):
    """ Credits:
    https://realpython.com/primer-on-python-decorators/#decorators-with-arguments
    """

    def decorator_trace(func):
        """ Print the function signature and return value.
        Optionally print the runtime (default: True)
        """

        @functools.wraps(func)
        def wrapper_debug(*args, **kwargs):
            lvl = logging.getLevelName(level)
            if not isinstance(lvl, int):
                lvl = logging.getLevelName("INFO")
                logging.warning("Unrecognised logging level!")
            args_repr = [repr(a) for a in args]
            kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            logger.log(level=lvl, msg=f"Calling {func.__name__}({signature})")
            start_time = time.perf_counter()
            value = func(*args, **kwargs)
            end_time = time.perf_counter()
            run_time = end_time - start_time
            if timeit:
                logger.log(
                    level=lvl, msg=f"Finished {func.__name__!r} in {run_time:.4f} secs"
                )
            logger.log(level=lvl, msg=f"{func.__name__!r} returned {value!r}")
            return value

        return wrapper_debug

    return decorator_trace



class RepoException(Exception):
    pass


#
class ShapeItException(Exception):
    def __init__(self, *args, **kwargs):
        self.output_method = kwargs.get("output_method")
        self.exit_code = 8
        if args:
            self.message = args[0]
        elif hasattr(self.__class__, "message"):
            self.message = self.__class__.message
        else:
            self.message = ""

    def __str__(self):
        return self.message


class NoAnswersError(ShapeItException):
    exit_code = 8


class NoStandardsError(Exception):
    pass


class AllowedTypesError(Exception):
    def __init__(self):
        self.message = "You have to use the specified types."
        super().__init__(self.message)

class NoRepoAccessError(Exception):
    def __init__(self):
        self.message = "You do not have access to the requested repo."
        super().__init__(self.message)

class NoGitRepo(Exception):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wommit', 'wommit.utils']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'emojis>=0.5.1,<0.6.0',
 'falcon>=2.0.0,<3.0.0',
 'gitpython>=3.1.3,<4.0.0',
 'keyring>=21.2.1,<22.0.0',
 'prompt_toolkit>=3.0.5,<4.0.0',
 'pycalver>=202007.36,<202008.0',
 'questionary>=1.5.2,<2.0.0',
 'requests>=2.24.0,<3.0.0',
 'webtest>=2.0.35,<3.0.0']

entry_points = \
{'console_scripts': ['commit = wommit.cli:c',
                     'shapeit = wommit.cli:cli',
                     'wmt = wommit.cli:c',
                     'wommit = wommit.cli:cli']}

setup_kwargs = {
    'name': 'wommit',
    'version': '20.8.38',
    'description': 'A tool for intuitively writing git commits that adhere to the Conventional Commits standards.',
    'long_description': "# Wommit\n\n**W**rite c**ommit** **t**ool\n\n---\n\n[![PyCalVer 20.08.0031-dev][version_img]][version_ref]\n[![PyPI Releases][pypi_img]][pypi_ref]\n\n#### A package for intuitively formatting appealing commmit messages with emojis, using an assortment of different methods.\n\n![CHECK ME OUT](https://i.imgur.com/VIXvQXY.png)\n\n---\n\n### Example usage:\n\n![EXAMPLE](https://i.imgur.com/EORqAkh.gif)\n\n---\n\n### Commands\n\n`wommit ...`:\n\n- `c`: Commit all staged files using an intuitive drop-down menu or a fast autocompletion prompt.\n\n  *Options:*\n  \n  - `-m`: Use menu mode, overriding default.\n\n  - `-e`: Use autocompletion mode, overriding default.\n  \n  - `-m [MESSAGE]`: Write a manual commit message, and commits if it's in the accepted format, as well as converting known types to emojis. \n  \n  - `-g`: Use global settings and data instead of local.\n\n  - `--test`:  Test either of the modes without commiting.\n  \n   - `--ic`:  Include staged files in the prompt display.\n  \n  \n- `check`: Manually check previously added commit. \n\n  *Options:*\n\n  - `-id [HASH]`: Check a commit message with the specified ID.\n  \n  - `-ids [HASH1] [HASH2]`: Check all commit messages between two IDs (newest ID first).\n\n  - `-m [MESSAGE]`:  Check if the given string passes the check.\n  \n  - `-l`:  Check all local commits that have not been pushed.\n\n- `configure ...`: Opens a prompt for adding/removing types and scopes.\n\n    - `e`: Edit current types and scopes.\n    \n    - `p`: Prints all types and scopes.\n    \n    - `s`: Edit settings.\n\n    *Options:*\n\n  - `-g`: Edit global settings.\n  \n  - `-t`: Manually test the functionality.\n  \n - `format`: Pastes the format a message needs to meet in order to pass the check.\n \n ---\n \n ## Info\n \n Use the [wommit-changelog-action](https://github.com/bkkp/wommit-changelog-action) in your wommit project to automatically release your project with appropriate changelogs.\n\n[version_img]: https://img.shields.io/static/v1.svg?label=Wommit&message=20.08.0031-dev&color=blue\n[version_ref]: https://pypi.org/project/wommit/\n[pypi_img]: https://img.shields.io/badge/PyPI-wheels-green.svg\n[pypi_ref]: https://pypi.org/project/wommit/\n\n",
    'author': 'Zylvian',
    'author_email': 'jovlisen@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bkkp/wommit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
